from diamond_miner.generators.database import probe_generator_from_database
from diamond_miner.generators.parallel import probe_generator_parallel
from diamond_miner.generators.standalone import probe_generator, probe_generator_by_flow

__all__ = (
    "probe_generator",
    "probe_generator_by_flow",
    "probe_generator_from_database",
    "probe_generator_parallel",
)



from .count import Count
from .count_rows import (
    CountFlowsPerPrefix,
    CountLinksPerPrefix,
    CountProbesPerPrefix,
    CountResultsPerPrefix,
)
from .create_flows_view import CreateFlowsView
from .create_links_table import CreateLinksTable
from .create_prefixes_table import CreatePrefixesTable
from .create_probes_table import CreateProbesTable
from .create_results_table import CreateResultsTable
from .create_tables import CreateTables
from .drop_tables import DropTables
from .get_invalid_prefixes import (
    GetInvalidPrefixes,
    GetPrefixesWithAmplification,
    GetPrefixesWithLoops,
)
from .get_links import GetLinks
from .get_links_from_view import GetLinksFromView
from .get_mda_probes import GetMDAProbes, InsertMDAProbes
from .get_nodes import GetNodes
from .get_prefixes import GetPrefixes
from .get_probes import GetProbes, GetProbesDiff
from .get_results import GetResults
from .get_sliding_prefixes import GetSlidingPrefixes
from .insert_links import InsertLinks
from .insert_prefixes import InsertPrefixes
from .query import (
    AddrType,
    FlowsQuery,
    LinksQuery,
    PrefixesQuery,
    ProbesQuery,
    Query,
    ResultsQuery,
    StoragePolicy,
    flows_table,
    links_table,
    prefixes_table,
    probes_table,
    results_table,
)

__all__ = (
    "AddrType",
    "Count",
    "CountFlowsPerPrefix",
    "CountLinksPerPrefix",
    "CountProbesPerPrefix",
    "CountResultsPerPrefix",
    "CreateFlowsView",
    "CreateLinksTable",
    "CreatePrefixesTable",
    "CreateProbesTable",
    "CreateResultsTable",
    "CreateTables",
    "DropTables",
    "GetLinks",
    "GetLinksFromView",
    "GetMDAProbes",
    "GetNodes",
    "GetPrefixes",
    "GetProbes",
    "GetProbesDiff",
    "GetResults",
    "GetSlidingPrefixes",
    "GetInvalidPrefixes",
    "GetPrefixesWithAmplification",
    "GetPrefixesWithLoops",
    "InsertMDAProbes",
    "InsertLinks",
    "InsertPrefixes",
    "Query",
    "FlowsQuery",
    "LinksQuery",
    "PrefixesQuery",
    "ProbesQuery",
    "ResultsQuery",
    "StoragePolicy",
    "flows_table",
    "links_table",
    "prefixes_table",
    "probes_table",
    "results_table",
)



from collections import namedtuple
from dataclasses import dataclass

from diamond_miner.defaults import DEFAULT_FAILURE_RATE, UNIVERSE_SUBSET
from diamond_miner.queries.query import LinksQuery, links_table, probes_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class GetMDAProbes(LinksQuery):
    adaptive_eps: bool = True
    dminer_lite: bool = True
    target_epsilon: float = DEFAULT_FAILURE_RATE

    Row = namedtuple("Row", "protocol,dst_prefix,cumulative_probes,ttls")

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        if self.adaptive_eps:
            eps_fragment = """
            arrayMax(links_per_ttl) AS max_links,
            if(max_links == 0, target_epsilon, 1 - exp(log(1 - target_epsilon) / max_links))
                AS epsilon,
            """
        else:
            eps_fragment = """
            target_epsilon AS epsilon,
            """

        if self.dminer_lite:
            dm_fragment = """
            arrayMap(k -> toUInt32(ceil(ln(epsilon / (k + 1)) / ln((k + 1 - 1) / (k + 1)))), links_per_ttl) AS mda_flows,
            """
        else:
            # TODO: Implement by computing Dh(v)
            raise NotImplementedError

        return f"""
        WITH
            {self.target_epsilon} AS target_epsilon,
            -- 1) Compute the links
            --  x.1       x.2        x.3
            -- (near_ttl, near_addr, far_addr)
            groupUniqArray((near_ttl, near_addr, far_addr)) AS links,
            -- 2) Count the number of links per TTL
            -- extract only the TTLs, this greatly speeds-up arrayCount
            arrayMap(x -> x.1, links) AS links_ttls,
            -- find the min/max TTLs
            -- we add +2 since range() is exclusive and that we compute the max over the *near* TTL
            range(arrayMin(links_ttls), arrayMax(links_ttls) + 2) AS TTLs,
            -- count distinct links per TTL
            arrayMap(t -> countEqual(links_ttls, t), TTLs) AS links_per_ttl,
            -- 3) Compute MDA stopping points
            {eps_fragment}
            -- 4) Compute the number of probes to send during the next round
            {dm_fragment}
            -- compute the number of probes to send during the next round
            -- => max of probes to send over TTL t and t-1
            arrayMap(i -> arrayMax([mda_flows[i], mda_flows[i - 1]]), arrayEnumerate(TTLs)) AS cumulative_probes
            -- TODO: Cleanup/optimize/rewrite/... below
            -- do not send probes to TTLs where no replies have been received
            -- it is unlikely that we will discover more at this TTL if the first 6 flows have seen nothing
            -- (see GetNextRoundStateless)
        SELECT
            probe_protocol,
            probe_dst_prefix,
            cumulative_probes,
            TTLs
        FROM {links_table(measurement_id)} AS links_table
        WHERE {self.filters(subset)}
        GROUP BY (probe_protocol, probe_src_addr, probe_dst_prefix)
        """


@dataclass(frozen=True)
class InsertMDAProbes(GetMDAProbes):
    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        assert self.round_leq
        return f"""
        INSERT INTO {probes_table(measurement_id)}
        WITH
            arrayJoin(arrayZip(TTLs, cumulative_probes)) AS ttl_probe
        SELECT
            probe_protocol,
            probe_dst_prefix,
            ttl_probe.1,
            ttl_probe.2,
            {self.round_leq + 1}
        FROM ({super().statement(measurement_id, subset)})
        """



from ipaddress import IPv6Address, IPv6Network, ip_network
from typing import Dict, List, Union

from diamond_miner.queries import (
    CountFlowsPerPrefix,
    CountLinksPerPrefix,
    CountProbesPerPrefix,
    CountResultsPerPrefix,
    FlowsQuery,
    LinksQuery,
    ProbesQuery,
    ResultsQuery,
)
from diamond_miner.utilities import common_parameters

Counts = Dict[IPv6Network, int]


async def subsets_for(
    query: Union[FlowsQuery, LinksQuery, ProbesQuery, ResultsQuery],
    url: str,
    measurement_id: str,
    max_items_per_subset: int = 8_000_000,
) -> List[IPv6Network]:
    if isinstance(query, FlowsQuery):
        count_query = CountFlowsPerPrefix(**common_parameters(query, FlowsQuery))
    elif isinstance(query, LinksQuery):
        count_query = CountLinksPerPrefix(**common_parameters(query, LinksQuery))  # type: ignore
    elif isinstance(query, ProbesQuery):
        count_query = CountProbesPerPrefix(**common_parameters(query, ProbesQuery))  # type: ignore
    elif isinstance(query, ResultsQuery):
        count_query = CountResultsPerPrefix(**common_parameters(query, ResultsQuery))  # type: ignore
    else:
        raise NotImplementedError
    counts = {
        addr_to_network(
            addr, count_query.prefix_len_v4, count_query.prefix_len_v6
        ): count
        for addr, count in await count_query.execute_async(url, measurement_id)
    }
    return split(counts, max_items_per_subset)


def split(counts: Counts, max_items_per_subset: int) -> List[IPv6Network]:
    """
    Return the IP networks such that there are no more than `max_items_per_subset`
    per network.

    :param counts: Number of items per prefix in the database table.
    :param max_items_per_subset: Maximum number of items per network.
    """
    candidates = [(ip_network("::/0"), n_items(counts, ip_network("::/0")))]
    subsets = []

    while candidates:
        candidate, n_replies = candidates.pop()
        if max_items_per_subset >= n_replies > 0:
            subsets.append(candidate)
        elif n_replies > 0:
            a, b = tuple(candidate.subnets(prefixlen_diff=1))
            n_items_a = n_items(counts, a)
            n_items_b = n_items(counts, b)
            if n_items_a + n_items_b == 0:
                subsets.append(candidate)
            else:
                candidates.append((a, n_items_a))
                candidates.append((b, n_items_b))

    return sorted(subsets)


def addr_to_network(
    addr: IPv6Address, prefix_len_v4: int, prefix_len_v6: int
) -> IPv6Network:
    if addr.ipv4_mapped:
        return IPv6Network(f"{addr}/{96+prefix_len_v4}")
    return IPv6Network(f"{addr}/{prefix_len_v6}")


def n_items(counts: Counts, subset: IPv6Network) -> int:
    total = 0
    for network, count_ in counts.items():
        if network.subnet_of(subset):
            total += count_
    return total



from dataclasses import dataclass
from typing import List

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries import GetInvalidPrefixes
from diamond_miner.queries.query import LinksQuery, links_table
from diamond_miner.typing import IPNetwork
from diamond_miner.utilities import common_parameters


@dataclass(frozen=True)
class GetLinks(LinksQuery):
    """
    Return the links pre-computed in the links table.

    >>> from diamond_miner.test import addr_to_string, url
    >>> links = GetLinks(filter_invalid_prefixes=False).execute(url, 'test_invalid_prefixes')
    >>> len(links)
    3
    >>> links = GetLinks(filter_invalid_prefixes=True).execute(url, 'test_invalid_prefixes')
    >>> len(links)
    1
    >>> links = GetLinks(include_metadata=False).execute(url, 'test_nsdi_example')
    >>> len(links)
    8
    >>> links = GetLinks(include_metadata=True).execute(url, 'test_nsdi_example')
    >>> len(links)
    8
    >>> links = GetLinks(near_or_far_addr="150.0.6.1").execute(url, 'test_nsdi_example')
    >>> len(links)
    3
    """

    filter_invalid_prefixes: bool = False
    "If true, exclude links from prefixes with amplification or loops."

    include_metadata: bool = False
    "If true, include the TTLs at which `near_addr` and `far_addr` were seen."

    def columns(self) -> List[str]:
        columns = [self._addr_cast("near_addr"), self._addr_cast("far_addr")]
        if self.include_metadata:
            columns = ["near_ttl", "far_ttl", *columns]
        return columns

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        if self.filter_invalid_prefixes:
            invalid_prefixes_query = GetInvalidPrefixes(
                **common_parameters(self, GetInvalidPrefixes)
            )
            prefix_filter = f"""
            probe_dst_prefix NOT IN ({invalid_prefixes_query.statement(measurement_id, subset)})
            """
        else:
            prefix_filter = "1"
        return f"""
        SELECT DISTINCT {','.join(self.columns())}
        FROM {links_table(measurement_id)}
        WHERE {self.filters(subset)} AND {prefix_filter}
        """



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import ResultsQuery, results_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class GetResults(ResultsQuery):
    """
    Return all the columns from the results table.

    >>> from diamond_miner.test import addr_to_string, url
    >>> rows = GetResults().execute(url, 'test_nsdi_example')
    >>> len(rows)
    85
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        SELECT *
        FROM {results_table(measurement_id)}
        WHERE {self.filters(subset)}
        """



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.fragments import cut_ipv6
from diamond_miner.queries.query import (
    FlowsQuery,
    LinksQuery,
    ProbesQuery,
    ResultsQuery,
    flows_table,
    links_table,
    probes_table,
    results_table,
)
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class CountFlowsPerPrefix(FlowsQuery):
    """
    Count rows per prefix.
    """

    prefix_len_v4: int = 8
    prefix_len_v6: int = 8

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        WITH {cut_ipv6('probe_dst_addr', self.prefix_len_v4, self.prefix_len_v6)} AS prefix
        SELECT prefix, count()
        FROM {flows_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY prefix
        """


@dataclass(frozen=True)
class CountLinksPerPrefix(LinksQuery):
    """
    Count rows per prefix.
    """

    prefix_len_v4: int = 8
    prefix_len_v6: int = 8

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        WITH {cut_ipv6('probe_dst_addr', self.prefix_len_v4, self.prefix_len_v6)} AS prefix
        SELECT prefix, count()
        FROM {links_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY prefix
        """


@dataclass(frozen=True)
class CountProbesPerPrefix(ProbesQuery):
    """
    Count the number of probes per prefix.
    """

    prefix_len_v4: int = 8
    prefix_len_v6: int = 8

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        assert self.round_eq
        return f"""
        WITH {cut_ipv6('probe_dst_prefix', self.prefix_len_v4, self.prefix_len_v6)} AS prefix
        SELECT prefix, sum(cumulative_probes)
        FROM {probes_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY prefix
        """


@dataclass(frozen=True)
class CountResultsPerPrefix(ResultsQuery):
    """
    Count rows per prefix.

    >>> from diamond_miner.test import addr_to_string, url
    >>> rows = CountResultsPerPrefix(prefix_len_v4=8, prefix_len_v6=8).execute(url, 'test_count_replies')
    >>> sorted((addr_to_string(a), b) for a, b in rows)
    [('1.0.0.0', 2), ('2.0.0.0', 1), ('204.0.0.0', 1)]
    """

    prefix_len_v4: int = 8
    prefix_len_v6: int = 8

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        WITH {cut_ipv6('probe_dst_addr', self.prefix_len_v4, self.prefix_len_v6)} AS prefix
        SELECT prefix, count()
        FROM {results_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY prefix
        """



import os
from typing import Iterable, Tuple

from diamond_miner.defaults import (
    DEFAULT_FAILURE_RATE,
    DEFAULT_PREFIX_LEN_V4,
    DEFAULT_PREFIX_LEN_V6,
    PROTOCOLS,
    UNIVERSE_SUBSET,
)
from diamond_miner.generators.standalone import split_prefix
from diamond_miner.queries import InsertMDAProbes
from diamond_miner.queries.query import client, probes_table
from diamond_miner.subsets import subsets_for
from diamond_miner.typing import IPNetwork


def insert_probe_counts(
    url: str,
    measurement_id: str,
    round_: int,
    # prefix (/32 or / 128 if nothing specified), protocol, ttls, n_probes
    prefixes: Iterable[Tuple[str, str, Iterable[int], int]],
    prefix_len_v4: int = DEFAULT_PREFIX_LEN_V4,
    prefix_len_v6: int = DEFAULT_PREFIX_LEN_V6,
    batch_size: int = 1_000_000,
) -> None:
    """
    >>> from diamond_miner.test import addr_to_string, create_tables, url
    >>> from diamond_miner.queries import GetProbes
    >>> create_tables(url, "test_probe_counts")
    >>> insert_probe_counts(url, "test_probe_counts", 1, [("8.8.0.0/22", "icmp", range(2, 5), 6)], batch_size=2)
    >>> rows = sorted(GetProbes(round_eq=1).execute(url, "test_probe_counts"))
    >>> len(rows)
    4
    >>> row = GetProbes.Row(*rows[0])
    >>> addr_to_string(row.dst_prefix)
    '8.8.0.0'
    >>> sorted(row.probes_per_ttl)
    [(2, 6), (3, 6), (4, 6)]
    >>> row = GetProbes.Row(*rows[1])
    >>> addr_to_string(row.dst_prefix)
    '8.8.1.0'
    >>> sorted(row.probes_per_ttl)
    [(2, 6), (3, 6), (4, 6)]
    """
    with client(url) as c:
        rows = []
        sql = f"INSERT INTO {probes_table(measurement_id)} VALUES"
        for prefix, protocol, ttls, n_probes in prefixes:
            protocol = PROTOCOLS[protocol]  # type: ignore
            for af, subprefix, subprefix_size in split_prefix(
                prefix, prefix_len_v4, prefix_len_v6
            ):
                for ttl in ttls:
                    rows.append((protocol, subprefix, ttl, n_probes, round_))
                if len(rows) >= batch_size:
                    c.execute(sql, rows)
                    rows.clear()
        c.execute(sql, rows)


def insert_mda_probe_counts(
    url: str,
    measurement_id: str,
    previous_round: int,
    adaptive_eps: bool = False,
    target_epsilon: float = DEFAULT_FAILURE_RATE,
    subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
) -> None:
    # TODO: set filter_partial and filter_virtual to false?
    InsertMDAProbes(
        adaptive_eps=adaptive_eps,
        round_leq=previous_round,
        filter_partial=True,
        filter_virtual=True,
        filter_inter_round=True,
        target_epsilon=target_epsilon,
    ).execute(url, measurement_id, subsets)


async def insert_mda_probe_counts_parallel(
    url: str,
    measurement_id: str,
    previous_round: int,
    adaptive_eps: bool = False,
    target_epsilon: float = DEFAULT_FAILURE_RATE,
    concurrent_requests: int = (os.cpu_count() or 2) // 2,
) -> None:
    # TODO: set filter_partial and filter_virtual to false?
    query = InsertMDAProbes(
        adaptive_eps=adaptive_eps,
        round_leq=previous_round,
        filter_partial=True,
        filter_virtual=True,
        filter_inter_round=True,
        target_epsilon=target_epsilon,
    )
    subsets = await subsets_for(query, url, measurement_id)
    await query.execute_concurrent(
        url, measurement_id, subsets, concurrent_requests=concurrent_requests
    )



from typing import Iterable, Iterator, Optional

from diamond_miner.defaults import (
    DEFAULT_PREFIX_SIZE_V4,
    DEFAULT_PREFIX_SIZE_V6,
    DEFAULT_PROBE_DST_PORT,
    DEFAULT_PROBE_SRC_PORT,
    PROTOCOLS,
    UNIVERSE_SUBSET,
)
from diamond_miner.logger import logger
from diamond_miner.mappers import SequentialFlowMapper
from diamond_miner.queries import GetProbesDiff
from diamond_miner.typing import FlowMapper, IPNetwork, Probe


def probe_generator_from_database(
    url: str,
    measurement_id: str,
    round_: int,
    *,
    mapper_v4: FlowMapper = SequentialFlowMapper(DEFAULT_PREFIX_SIZE_V4),
    mapper_v6: FlowMapper = SequentialFlowMapper(DEFAULT_PREFIX_SIZE_V6),
    probe_src_port: int = DEFAULT_PROBE_SRC_PORT,
    probe_dst_port: int = DEFAULT_PROBE_DST_PORT,
    probe_ttl_geq: Optional[int] = None,
    probe_ttl_leq: Optional[int] = None,
    subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
) -> Iterator[Probe]:
    """
    TODO: Doctest, note that this doesn't randomize probes.
    >>> from ipaddress import ip_address
    >>> from diamond_miner.insert import insert_probe_counts
    >>> from diamond_miner.test import create_tables, url
    >>> create_tables(url, "test_probe_gen")
    >>> insert_probe_counts(url, "test_probe_gen", 1, [("8.8.0.0/23", "icmp", [1, 2], 2)])
    >>> probes = list(probe_generator_from_database(url, "test_probe_gen", 1))
    >>> len(probes)
    8
    >>> (str(ip_address(probes[0][0])), *probes[0][1:])
    ('::ffff:808:100', 24000, 33434, 1, 'icmp')
    """
    rows = GetProbesDiff(
        round_eq=round_, probe_ttl_geq=probe_ttl_geq, probe_ttl_leq=probe_ttl_leq
    ).execute_iter(url, measurement_id, subsets)
    for row in rows:
        row = GetProbesDiff.Row(*row)

        dst_prefix_int = int(row.dst_prefix)
        mapper = mapper_v4 if row.dst_prefix.ipv4_mapped else mapper_v6
        protocol_str = PROTOCOLS[row.protocol]

        for ttl, total_probes, already_sent in row.probes_per_ttl:
            for flow_id in range(already_sent, total_probes):
                addr_offset, port_offset = mapper.offset(flow_id, dst_prefix_int)
                dst_addr = dst_prefix_int + addr_offset
                src_port = probe_src_port + port_offset
                if src_port > (2 ** 16 - 1):
                    # TEMP: Log prefixes that overflows the port number and skip prefix.
                    logger.warning("Port overflow for %s", row)
                    break
                yield dst_addr, src_port, probe_dst_port, ttl, protocol_str  # type: ignore



from dataclasses import asdict, dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries import GetLinksFromView
from diamond_miner.queries.query import FlowsQuery, links_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class InsertLinks(FlowsQuery):
    """Create the tables necessary for a measurement."""

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        links_query = GetLinksFromView(**asdict(self)).statement(measurement_id, subset)
        return f"""
        INSERT INTO {links_table(measurement_id)}
        SELECT * FROM ({links_query})
        """



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import ResultsQuery, results_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class GetSlidingPrefixes(ResultsQuery):
    """
    Get the prefixes to probe for a given sliding window.
    """

    stopping_condition: int = 0  # number of stars
    window_max_ttl: int = 0  # set to 0 to return every prefix

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        assert self.filter_destination_host and self.time_exceeded_only
        return f"""
        SELECT
            probe_protocol,
            {self._addr_cast('probe_src_addr')},
            {self._addr_cast('probe_dst_prefix')}
        FROM {results_table(measurement_id)}
        WHERE {self.filters(subset)}
        AND probe_ttl >= {self.window_max_ttl - self.stopping_condition}
        GROUP BY (
            probe_protocol,
            probe_src_addr,
            probe_dst_prefix,
        )
        """
